@@ -1,4 +1,4 @@
-/* $Id: unicode_info.c,v 1.13 2003/03/11 05:40:31 nlevitt Exp $ */
+/* $Id: unicode_info.c,v 1.16 2003/03/29 21:24:51 nlevitt Exp $ */
 /*
  * Copyright (c) 2002  Noah Levitt <nlevitt@users.sourceforge.net>
  *
@@ -23,14 +23,75 @@
 
 #include <gtk/gtk.h>
 #include <string.h>
+#include <gucharmap_intl.h>
 #include <unicode_info.h>
 
+
+typedef struct 
+{
+  gunichar index;
+  const gchar *name;
+} 
+UnicodeData;
+
+
+typedef struct 
+{
+  gunichar index;
+  const gchar *kDefinition;
+  const gchar *kCantonese;
+  const gchar *kMandarin;
+  const gchar *kTang;
+  const gchar *kKorean;
+  const gchar *kJapeneseKun;
+  const gchar *kJapaneseOn;
+} 
+Unihan;
+
+
+
+typedef struct 
+{
+  gunichar index;
+  gchar *value;
+}
+UnicharString;
+
+typedef struct
+{
+  gunichar index;
+  gunichar value;
+}
+UnicharUnichar;
+
+typedef struct
+{
+  gunichar index;
+  gint equals_index;  /* -1 means */
+  gint stars_index;   /* this character */
+  gint exes_index;    /* doesn't */
+  gint pounds_index;  /* have any */
+  gint colons_index;
+}
+NamesList;
+
+
+typedef struct
+{
+  gunichar first;
+  gunichar last;
+  GUnicodeType category;
+}
+UnicodeCategory;
+
+
 #include <unicode/unicode_data.cI>
 #include <unicode/unicode_blocks.cI>
 #if ENABLE_UNIHAN
 # include <unicode/unicode_unihan.cI>
 #endif
 #include <unicode/unicode_nameslist.cI>
+#include <unicode/unicode_categories.cI>
 
 
 /* constants for hangul (de)composition, see UAX #15 */
@@ -62,7 +123,7 @@ static const gchar * const JAMO_T_TABLE[] = {
 };
 
 /* computes the hangul name as per UAX #15 */
-const gchar *
+G_CONST_RETURN gchar *
 get_hangul_syllable_name (gunichar s)
 {
   static gchar buf[32];
@@ -76,82 +137,82 @@ get_hangul_syllable_name (gunichar s)
   VIndex = (SIndex % NCount) / TCount;
   TIndex = SIndex % TCount;
 
-  g_snprintf (buf, 32, "HANGUL SYLLABLE %s%s%s", JAMO_L_TABLE[LIndex],
+  g_snprintf (buf, 32, _("HANGUL SYLLABLE %s%s%s"), JAMO_L_TABLE[LIndex],
               JAMO_V_TABLE[VIndex], JAMO_T_TABLE[TIndex]);
 
   return buf;
 }
 
 
-const gchar *
+G_CONST_RETURN gchar *
 get_unicode_name (gunichar uc)
 {
   if (uc >= 0x3400 && uc <= 0x4DB5)
-    return "<CJK Ideograph Extension A>";
+    return _("<CJK Ideograph Extension A>");
   else if (uc >= 0x4e00 && uc <= 0x9fa5)
-    return "<CJK Ideograph>";
+    return _("<CJK Ideograph>");
   else if (uc >= 0xac00 && uc <= 0xd7af)
     return get_hangul_syllable_name (uc);
   else if (uc >= 0xD800 && uc <= 0xDB7F) 
-    return "<Non Private Use High Surrogate>";
+    return _("<Non Private Use High Surrogate>");
   else if (uc >= 0xDB80 && uc <= 0xDBFF) 
-    return "<Private Use High Surrogate>";
+    return _("<Private Use High Surrogate>");
   else if (uc >= 0xDC00 && uc <= 0xDFFF)
-    return "<Low Surrogate, Last>";
+    return _("<Low Surrogate, Last>");
   else if (uc >= 0xE000 && uc <= 0xF8FF) 
-    return "<Private Use>";
+    return _("<Private Use>");
   else if (uc >= 0xF0000 && uc <= 0xFFFFD)
-    return "<Plane 15 Private Use>";
+    return _("<Plane 15 Private Use>");
   else if (uc >= 0x100000 && uc <= 0x10FFFD)
-    return "<Plane 16 Private Use>";
+    return _("<Plane 16 Private Use>");
   else if (uc >= 0x20000 && uc <= 0x2A6D6)
-    return "<CJK Ideograph Extension B>";
+    return _("<CJK Ideograph Extension B>");
   else
     {
       const gchar *x = get_unicode_data_name (uc);
       if (x == NULL)
-        return "<not assigned>";
+        return _("<not assigned>");
       else
         return x;
     }
 }
 
 
-const gchar *
+G_CONST_RETURN gchar *
 get_unicode_category_name (gunichar uc)
 {
-  switch (g_unichar_type (uc))
+  switch (unichar_type (uc))
     {
-      case G_UNICODE_CONTROL: return "Other, Control";
-      case G_UNICODE_FORMAT: return "Other, Format";
-      case G_UNICODE_UNASSIGNED: return "Other, Not Assigned";
-      case G_UNICODE_PRIVATE_USE: return "Other, Private Use";
-      case G_UNICODE_SURROGATE: return "Other, Surrogate";
-      case G_UNICODE_LOWERCASE_LETTER: return "Letter, Lowercase";
-      case G_UNICODE_MODIFIER_LETTER: return "Letter, Modifier";
-      case G_UNICODE_OTHER_LETTER: return "Letter, Other";
-      case G_UNICODE_TITLECASE_LETTER: return "Letter, Titlecase";
-      case G_UNICODE_UPPERCASE_LETTER: return "Letter, Uppercase";
-      case G_UNICODE_COMBINING_MARK: return "Mark, Spacing Combining";
-      case G_UNICODE_ENCLOSING_MARK: return "Mark, Enclosing";
-      case G_UNICODE_NON_SPACING_MARK: return "Mark, Non-Spacing";
-      case G_UNICODE_DECIMAL_NUMBER: return "Number, Decimal Digit";
-      case G_UNICODE_LETTER_NUMBER: return "Number, Letter";
-      case G_UNICODE_OTHER_NUMBER: return "Number, Other";
-      case G_UNICODE_CONNECT_PUNCTUATION: return "Punctuation, Connector";
-      case G_UNICODE_DASH_PUNCTUATION: return "Punctuation, Dash";
-      case G_UNICODE_CLOSE_PUNCTUATION: return "Punctuation, Close";
-      case G_UNICODE_FINAL_PUNCTUATION: return "Punctuation, Final quote ";
-      case G_UNICODE_INITIAL_PUNCTUATION: return "Punctuation, Initial quote";
-      case G_UNICODE_OTHER_PUNCTUATION: return "Punctuation, Other";
-      case G_UNICODE_OPEN_PUNCTUATION: return "Punctuation, Open";
-      case G_UNICODE_CURRENCY_SYMBOL: return "Symbol, Currency";
-      case G_UNICODE_MODIFIER_SYMBOL: return "Symbol, Modifier";
-      case G_UNICODE_MATH_SYMBOL: return "Symbol, Math";
-      case G_UNICODE_OTHER_SYMBOL: return "Symbol, Other";
-      case G_UNICODE_LINE_SEPARATOR: return "Separator, Line";
-      case G_UNICODE_PARAGRAPH_SEPARATOR: return "Separator, Paragraph";
-      case G_UNICODE_SPACE_SEPARATOR: return "Separator, Space";
+      case G_UNICODE_CONTROL: return _("Other, Control");
+      case G_UNICODE_FORMAT: return _("Other, Format");
+      case G_UNICODE_UNASSIGNED: return _("Other, Not Assigned");
+      case G_UNICODE_PRIVATE_USE: return _("Other, Private Use");
+      case G_UNICODE_SURROGATE: return _("Other, Surrogate");
+      case G_UNICODE_LOWERCASE_LETTER: return _("Letter, Lowercase");
+      case G_UNICODE_MODIFIER_LETTER: return _("Letter, Modifier");
+      case G_UNICODE_OTHER_LETTER: return _("Letter, Other");
+      case G_UNICODE_TITLECASE_LETTER: return _("Letter, Titlecase");
+      case G_UNICODE_UPPERCASE_LETTER: return _("Letter, Uppercase");
+      case G_UNICODE_COMBINING_MARK: return _("Mark, Spacing Combining");
+      case G_UNICODE_ENCLOSING_MARK: return _("Mark, Enclosing");
+      case G_UNICODE_NON_SPACING_MARK: return _("Mark, Non-Spacing");
+      case G_UNICODE_DECIMAL_NUMBER: return _("Number, Decimal Digit");
+      case G_UNICODE_LETTER_NUMBER: return _("Number, Letter");
+      case G_UNICODE_OTHER_NUMBER: return _("Number, Other");
+      case G_UNICODE_CONNECT_PUNCTUATION: return _("Punctuation, Connector");
+      case G_UNICODE_DASH_PUNCTUATION: return _("Punctuation, Dash");
+      case G_UNICODE_CLOSE_PUNCTUATION: return _("Punctuation, Close");
+      case G_UNICODE_FINAL_PUNCTUATION: return _("Punctuation, Final Quote");
+      case G_UNICODE_INITIAL_PUNCTUATION: return _("Punctuation, Initial Quote");
+      case G_UNICODE_OTHER_PUNCTUATION: return _("Punctuation, Other");
+      case G_UNICODE_OPEN_PUNCTUATION: return _("Punctuation, Open");
+      case G_UNICODE_CURRENCY_SYMBOL: return _("Symbol, Currency");
+      case G_UNICODE_MODIFIER_SYMBOL: return _("Symbol, Modifier");
+      case G_UNICODE_MATH_SYMBOL: return _("Symbol, Math");
+      case G_UNICODE_OTHER_SYMBOL: return _("Symbol, Other");
+      case G_UNICODE_LINE_SEPARATOR: return _("Separator, Line");
+      case G_UNICODE_PARAGRAPH_SEPARATOR: return _("Separator, Paragraph");
+      case G_UNICODE_SPACE_SEPARATOR: return _("Separator, Space");
       default: return "";
     }
 }
@@ -229,7 +290,7 @@ unicode_canonical_decomposition (gunichar ch, gsize   *result_len)
 
 
 /* does a binary search on unicode_data */
-const gchar *
+G_CONST_RETURN gchar *
 get_unicode_data_name (gunichar uc)
 {
   gint min = 0;
@@ -255,7 +316,7 @@ get_unicode_data_name (gunichar uc)
 
 
 /* ascii case-insensitive substring search (source ripped from glib) */
-static const gchar *
+static G_CONST_RETURN gchar *
 ascii_case_strrstr (const gchar *haystack, const gchar *needle)
 {
   gsize i;
@@ -294,21 +355,25 @@ ascii_case_strrstr (const gchar *haystack, const gchar *needle)
 
 
 /* case insensitive; returns (gunichar)(-1) if nothing found */
+/* direction must be +1 or -1 */
 gunichar
-find_next_substring_match (gunichar start, gunichar unichar_max,
-                           const gchar *search_text)
+find_substring_match (gunichar start, 
+                      const gchar *search_text,
+                      gint direction)
 {
-  gint min = 0;
-  gint mid = 0;
   gint max = sizeof (unicode_data) / sizeof (UnicodeData) - 1;
   gint i0;
   gint i;
 
+  g_assert (direction == -1 || direction == 1);
+
   /* locate the start character by binary search */
-  if (start < unicode_data[0].index || start > unichar_max)
+  if (start < unicode_data[0].index || start > UNICHAR_MAX)
     i0 = 0;
   else
     {
+      gint min = 0, mid = 0;
+
       while (max >= min) 
         {
           mid = (min + max) / 2;
@@ -323,16 +388,24 @@ find_next_substring_match (gunichar start, gunichar unichar_max,
       i0 = mid;
     }
 
+  max = sizeof (unicode_data) / sizeof (UnicodeData) - 1;
   /* try substring match on each */
-  max = sizeof (unicode_data) / sizeof (UnicodeData);
-  for (i = i0+1;  i != i0;  )
+  for (i = i0 + direction;  i != i0;  )
     {
+      if (unicode_data[i].index > UNICHAR_MAX)
+        {
+          i += direction;
+          continue;
+        }
+
       if (ascii_case_strrstr (unicode_data[i].name, search_text) != NULL)
         return unicode_data[i].index;
 
-      i++;
-      if (i >= max || unicode_data[i].index > unichar_max)
+      i += direction;
+      if (i > max)
         i = 0;
+      else if (i < 0)
+        i = max;
     }
 
   /* if the start character matches we want to return a match */
@@ -347,7 +420,7 @@ find_next_substring_match (gunichar start, gunichar unichar_max,
 
 /* does a binary search; also caches most recent, since it will often be
  * called in succession on the same character */
-static const Unihan *
+static G_CONST_RETURN Unihan *
 _get_unihan (gunichar uc)
 {
   static gunichar most_recent_searched;
@@ -383,7 +456,7 @@ _get_unihan (gunichar uc)
 }
 
 
-const gchar * 
+G_CONST_RETURN gchar * 
 get_unicode_kDefinition (gunichar uc)
 {
   const Unihan *uh = _get_unihan (uc);
@@ -393,7 +466,7 @@ get_unicode_kDefinition (gunichar uc)
     return uh->kDefinition;
 }
 
-const gchar * 
+G_CONST_RETURN gchar * 
 get_unicode_kCantonese (gunichar uc)
 {
   const Unihan *uh = _get_unihan (uc);
@@ -403,7 +476,7 @@ get_unicode_kCantonese (gunichar uc)
     return uh->kCantonese;
 }
 
-const gchar * 
+G_CONST_RETURN gchar * 
 get_unicode_kMandarin (gunichar uc)
 {
   const Unihan *uh = _get_unihan (uc);
@@ -413,7 +486,7 @@ get_unicode_kMandarin (gunichar uc)
     return uh->kMandarin;
 }
 
-const gchar * 
+G_CONST_RETURN gchar * 
 get_unicode_kTang (gunichar uc)
 {
   const Unihan *uh = _get_unihan (uc);
@@ -423,7 +496,7 @@ get_unicode_kTang (gunichar uc)
     return uh->kTang;
 }
 
-const gchar * 
+G_CONST_RETURN gchar * 
 get_unicode_kKorean (gunichar uc)
 {
   const Unihan *uh = _get_unihan (uc);
@@ -433,7 +506,7 @@ get_unicode_kKorean (gunichar uc)
     return uh->kKorean;
 }
 
-const gchar * 
+G_CONST_RETURN gchar * 
 get_unicode_kJapaneseKun (gunichar uc)
 {
   const Unihan *uh = _get_unihan (uc);
@@ -443,7 +516,7 @@ get_unicode_kJapaneseKun (gunichar uc)
     return uh->kJapeneseKun;
 }
 
-const gchar * 
+G_CONST_RETURN gchar * 
 get_unicode_kJapaneseOn (gunichar uc)
 {
   const Unihan *uh = _get_unihan (uc);
@@ -455,43 +528,43 @@ get_unicode_kJapaneseOn (gunichar uc)
 
 #else /* #if ENABLE_UNIHAN */
 
-const gchar * 
+G_CONST_RETURN gchar * 
 get_unicode_kDefinition (gunichar uc)
 {
   return "This feature was not compiled in.";
 }
 
-const gchar * 
+G_CONST_RETURN gchar * 
 get_unicode_kCantonese (gunichar uc)
 {
   return "This feature was not compiled in.";
 }
 
-const gchar * 
+G_CONST_RETURN gchar * 
 get_unicode_kMandarin (gunichar uc)
 {
   return "This feature was not compiled in.";
 }
 
-const gchar * 
+G_CONST_RETURN gchar * 
 get_unicode_kTang (gunichar uc)
 {
   return "This feature was not compiled in.";
 }
 
-const gchar * 
+G_CONST_RETURN gchar * 
 get_unicode_kKorean (gunichar uc)
 {
   return "This feature was not compiled in.";
 }
 
-const gchar * 
+G_CONST_RETURN gchar * 
 get_unicode_kJapaneseKun (gunichar uc)
 {
   return "This feature was not compiled in.";
 }
 
-const gchar * 
+G_CONST_RETURN gchar * 
 get_unicode_kJapaneseOn (gunichar uc)
 {
   return "This feature was not compiled in.";
@@ -502,7 +575,7 @@ get_unicode_kJapaneseOn (gunichar uc)
 
 /* does a binary search; also caches most recent, since it will often be
  * called in succession on the same character */
-static const NamesList *
+static G_CONST_RETURN NamesList *
 get_nameslist (gunichar uc)
 {
   static gunichar most_recent_searched;
@@ -566,7 +639,7 @@ get_nameslist_exes (gunichar uc)
 
 /* returns newly allocated null-terminated array of gchar* */
 /* the items are const, but the array should be freed by the caller */
-const gchar **
+G_CONST_RETURN gchar **
 get_nameslist_equals (gunichar uc)
 {
   const NamesList *nl;
@@ -593,7 +666,7 @@ get_nameslist_equals (gunichar uc)
 
 /* returns newly allocated null-terminated array of gchar* */
 /* the items are const, but the array should be freed by the caller */
-const gchar **
+G_CONST_RETURN gchar **
 get_nameslist_stars (gunichar uc)
 {
   const NamesList *nl;
@@ -620,7 +693,7 @@ get_nameslist_stars (gunichar uc)
 
 /* returns newly allocated null-terminated array of gchar* */
 /* the items are const, but the array should be freed by the caller */
-const gchar **
+G_CONST_RETURN gchar **
 get_nameslist_pounds (gunichar uc)
 {
   const NamesList *nl;
@@ -647,7 +720,7 @@ get_nameslist_pounds (gunichar uc)
 
 /* returns newly allocated null-terminated array of gchar* */
 /* the items are const, but the array should be freed by the caller */
-const gchar **
+G_CONST_RETURN gchar **
 get_nameslist_colons (gunichar uc)
 {
   const NamesList *nl;
@@ -671,3 +744,147 @@ get_nameslist_colons (gunichar uc)
   return colons;
 }
 
+
+#define UNICODE_VALID(Char)                   \
+    ((Char) < 0x110000 &&                     \
+    ((Char) < 0xD800 || (Char) >= 0xE000) &&  \
+    ((Char) < 0xFDD0 || (Char) > 0xFDEF) &&   \
+    ((Char) & 0xFFFF) != 0xFFFE && ((Char) & 0xFFFF) != 0xFFFF)
+
+/* an up-to-date replacement for g_unichar_validate */
+gboolean
+unichar_validate (gunichar ch)
+{
+  return UNICODE_VALID (ch);
+}
+
+
+/**
+ * unichar_to_printable_utf8
+ * @uc: a unicode character 
+ * @outbuf: output buffer, must have at least 10 bytes of space.
+ *          If %NULL, the length will be computed and returned
+ *          and nothing will be written to @outbuf.
+ *
+ * Converts a single character to UTF-8 suitable for rendering. Check the
+ * source to see what this means. ;-)
+ * 
+ *
+ * Return value: number of bytes written
+ **/
+gint
+unichar_to_printable_utf8 (gunichar uc, gchar *outbuf)
+{
+  /* Unicode Standard 3.2, section 2.6, "By convention, diacritical marks
+   * used by the Unicode Standard may be exhibited in (apparent) isolation
+   * by applying them to U+0020 SPACE or to U+00A0 NO BREAK SPACE." */
+
+  /* 17:10 < owen> noah: I'm *not* claiming that what Pango does currently
+   *               is right, but convention isn't a requirement. I think
+   *               it's probably better to do the Uniscribe thing and put
+   *               the lone combining mark on a dummy character and require
+   *               ZWJ
+   * 17:11 < noah> owen: do you mean that i should put a ZWJ in there, or
+   *               that pango will do that?
+   * 17:11 < owen> noah: I mean, you should (assuming some future more
+   *               capable version of Pango) put it in there
+   */
+
+  if (! unichar_validate (uc) || (! unichar_isgraph (uc) 
+      && unichar_type (uc) != G_UNICODE_PRIVATE_USE))
+    return 0;
+  else if (unichar_type (uc) == G_UNICODE_COMBINING_MARK
+      || unichar_type (uc) == G_UNICODE_ENCLOSING_MARK
+      || unichar_type (uc) == G_UNICODE_NON_SPACING_MARK)
+    {
+      gint x;
+
+      outbuf[0] = ' ';
+      outbuf[1] = '\xe2'; /* ZERO */ 
+      outbuf[2] = '\x80'; /* WIDTH */
+      outbuf[3] = '\x8d'; /* JOINER (0x200D) */
+
+      x = g_unichar_to_utf8 (uc, outbuf + 4);
+
+      return x + 4;
+    }
+  else
+    return g_unichar_to_utf8 (uc, outbuf);
+}
+
+
+/**
+ * unichar_type:
+ * @c: a Unicode character
+ * 
+ * Classifies a Unicode character by type.
+ * 
+ * Return value: the type of the character.
+ **/
+GUnicodeType
+unichar_type (gunichar uc)
+{
+  gint min = 0;
+  gint mid;
+  gint max = sizeof (unicode_categories) / sizeof (UnicodeCategory) - 1;
+
+  if (uc < unicode_categories[0].first || uc > unicode_categories[max].last)
+    return G_UNICODE_UNASSIGNED;
+
+  while (max >= min) 
+    {
+      mid = (min + max) / 2;
+      if (uc > unicode_categories[mid].last)
+        min = mid + 1;
+      else if (uc < unicode_categories[mid].first)
+        max = mid - 1;
+      else
+        return unicode_categories[mid].category;
+    }
+
+  return G_UNICODE_UNASSIGNED;
+}
+
+
+/**
+ * unichar_isdefined:
+ * @uc: a Unicode character
+ * 
+ * Determines if a given character is assigned in the Unicode
+ * standard.
+ *
+ * Return value: %TRUE if the character has an assigned value
+ **/
+gboolean
+unichar_isdefined (gunichar uc)
+{
+  return unichar_type (uc) != G_UNICODE_UNASSIGNED;
+}
+
+
+/**
+ * unichar_isgraph:
+ * @uc: a Unicode character
+ * 
+ * Determines whether a character is printable and not a space
+ * (returns %FALSE for control characters, format characters, and
+ * spaces). g_unichar_isprint() is similar, but returns %TRUE for
+ * spaces. Given some UTF-8 text, obtain a character value with
+ * g_utf8_get_char().
+ * 
+ * Return value: %TRUE if @c is printable unless it's a space
+ **/
+gboolean
+unichar_isgraph (gunichar uc)
+{
+  GUnicodeType t = unichar_type (uc);
+
+  return (t != G_UNICODE_CONTROL
+          && t != G_UNICODE_FORMAT
+          && t != G_UNICODE_UNASSIGNED
+          && t != G_UNICODE_PRIVATE_USE
+          && t != G_UNICODE_SURROGATE
+          && t != G_UNICODE_SPACE_SEPARATOR);
+}
+
+
